# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Factory functions for creating proximal operators.

Functions with ``convex_conj`` mean the proximal of the convex conjugate and
are provided for convenience.

For more details see :ref:`proximal_operators` and references therein. For
more details on proximal operators including how to evaluate the proximal
operator of a variety of functions see [PB2014].

References
----------
[PB2014] Parikh, N, and Boyd, S. *Proximal Algorithms*.
Foundations and Trends in Optimization, 1 (2014), pp 127-239.
"""

from __future__ import print_function, division, absolute_import

import warnings

import numpy as np
import math
from odl.core.operator import (
    Operator, IdentityOperator, ConstantOperator, DiagonalOperator,
    PointwiseNorm, MultiplyOperator)
from odl.core.space.pspace import ProductSpace, ProductSpaceElement
from odl.core.space.base_tensors import Tensor
from odl.core.set.space import LinearSpace, LinearSpaceElement
from odl.core.array_API_support.element_wise import maximum, minimum, abs, divide, sign, square, sqrt, less_equal, logical_not, exp
from odl.core.array_API_support.statistical import sum
from odl.core.util.scipy_compatibility import lambertw, scipy_lambertw
from odl.core.util.dtype_utils import is_complex_dtype


__all__ = ('combine_proximals', 'proximal_convex_conj', 'proximal_translation',
           'proximal_arg_scaling', 'proximal_quadratic_perturbation',
           'proximal_composition', 'proximal_const_func',
           'proximal_box_constraint', 'proximal_nonnegativity',
           'proximal_l1', 'proximal_convex_conj_l1',
           'proximal_l2', 'proximal_convex_conj_l2',
           'proximal_linfty', 'proximal_convex_conj_linfty',
           'proj_simplex', 'proj_l1',
           'proximal_l2_squared', 'proximal_convex_conj_l2_squared',
           'proximal_l1_l2', 'proximal_convex_conj_l1_l2',
           'proximal_convex_conj_kl', 'proximal_convex_conj_kl_cross_entropy',
           'proximal_huber')

def _numerical_epsilon(space: LinearSpace):
    """Determine numerical precision, preferably from the data type of the space,
    else defaulting to double-precision float in case there is no single dtype
    (e.g. `ProductSpaces`)."""
    dtype_id = getattr(space, 'dtype_identifier', 'float64')
    # Always use NumPy for resolution, assuming that different backends would offer
    # the same precision for the corresponding types.
    eps = np.finfo(dtype_id).resolution * 10
    return eps

def combine_proximals(*factory_list):
    r"""Combine proximal operators into a diagonal product space operator.

    This assumes the functional to be separable across variables in order to
    make use of the separable sum property of proximal operators.

    Parameters
    ----------
    factory_list : sequence of callables
        Proximal operator factories to be combined.

    Returns
    -------
    diag_op : function
        Returns a diagonal product space operator factory to be initialized
        with the same step size parameter

    Notes
    -----
    That two functionals :math:`F` and :math:`G` are separable across variables
    means that :math:`F((x, y)) = F(x)` and :math:`G((x, y)) = G(y)`, and in
    this case the proximal operator of the sum is given by

    .. math::
        \mathrm{prox}_{\sigma (F(x) + G(y))}(x, y) =
        (\mathrm{prox}_{\sigma F}(x), \mathrm{prox}_{\sigma G}(y)).
    """
    def diag_op_factory(sigma):
        """Diagonal matrix of operators.

        Parameters
        ----------
        sigma : positive float or sequence of positive floats
            Step size parameter(s), if a sequence, the length must match
            the length of the ``factory_list``.

        Returns
        -------
        diag_op : `DiagonalOperator`
        """
        if np.isscalar(sigma):
            sigma = [sigma] * len(factory_list)

        return DiagonalOperator(
            *[factory(sigmai)
              for sigmai, factory in zip(sigma, factory_list)])

    return diag_op_factory


def proximal_convex_conj(prox_factory):
    r"""Calculate the proximal of the dual using Moreau decomposition.

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The Moreau identity states that for any convex function :math:`F` with
    convex conjugate :math:`F^*`, the proximals satisfy

    .. math::
        \mathrm{prox}_{\sigma F^*}(x) +\sigma \,
        \mathrm{prox}_{F / \sigma}(x / \sigma) = x

    where :math:`\sigma` is a scalar step size. Using this, the proximal of
    the convex conjugate is given by

    .. math::
        \mathrm{prox}_{\sigma F^*}(x) =
        x - \sigma \, \mathrm{prox}_{F / \sigma}(x / \sigma)

    Note that since :math:`(F^*)^* = F`, this can be used to get the proximal
    of the original function from the proximal of the convex conjugate.

    For reference on the Moreau identity, see [CP2011c].

    References
    ----------
    [CP2011c] Combettes, P L, and Pesquet, J-C. *Proximal splitting
    methods in signal processing.* In:  Bauschke, H H, Burachik, R S,
    Combettes, P L, Elser, V, Luke, D R, and Wolkowicz, H. Fixed-point
    algorithms for inverse problems in science and engineering, Springer,
    2011.
    """
    def convex_conj_prox_factory(sigma):
        """Create proximal for the dual with a given sigma.

        Parameters
        ----------
        sigma : positive float or array-like
            Step size parameter. Can be a pointwise positive space element or
            a sequence of positive floats if `prox_factory` supports that.

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``s * F^*`` where ``s`` is the step size
        """

        # Get the underlying space. At the same time, check if the given
        # prox_factory accepts stepsize objects of the type given by sigma.
        space = prox_factory(sigma).domain

        mult_inner = MultiplyOperator(1.0 / sigma, domain=space, range=space)
        mult_outer = MultiplyOperator(sigma, domain=space, range=space)
        result = (IdentityOperator(space) -
                  mult_outer * prox_factory(1.0 / sigma) * mult_inner)
        return result

    return convex_conj_prox_factory


def proximal_translation(prox_factory, y):
    r"""Calculate the proximal of the translated function F(x - y).

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``.
    y : Element in domain of ``F``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a functional :math:`F`, this is calculated according to the rule

    .. math::
        \mathrm{prox}_{\sigma F( \cdot - y)}(x) =
        y + \mathrm{prox}_{\sigma F}(x - y)

    where :math:`y` is the translation, and :math:`\sigma` is the step size.

    For reference on the identity used, see [CP2011c].

    References
    ----------
    [CP2011c] Combettes, P L, and Pesquet, J-C. *Proximal splitting
    methods in signal processing.* In:  Bauschke, H H, Burachik, R S,
    Combettes, P L, Elser, V, Luke, D R, and Wolkowicz, H. Fixed-point
    algorithms for inverse problems in science and engineering, Springer,
    2011.
    """

    def translation_prox_factory(sigma):
        """Create proximal for the translation with a given sigma.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``s * F( . - y)`` where ``s`` is the
            step size
        """
        return (ConstantOperator(y) + prox_factory(sigma) *
                (IdentityOperator(y.space) - ConstantOperator(y)))

    return translation_prox_factory


def proximal_arg_scaling(prox_factory, scaling):
    r"""Calculate the proximal of function F(x * scaling).

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``
    scaling : float or sequence of floats or space element
        Scaling parameter. The permissible types depent on the stepsizes
        accepted by prox_factory. It may not contain any nonzero imaginary
        parts. If it is a scalar, it may be zero, in which case the
        resulting proxmial operator is the identity. If not a scalar,
        it may not contain any zero components.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a functional :math:`F`, and scaling factor :math:`\alpha` this is
    calculated according to the rule

    .. math::
        \mathrm{prox}_{\sigma F(\alpha \, \cdot)}(x) =
        \frac{1}{\alpha}
        \mathrm{prox}_{\sigma \alpha^2 F(\cdot) }(\alpha x)

    where :math:`\sigma` is the step size.

    For reference on the identity used, see [CP2011c].

    References
    ----------
    [CP2011c] Combettes, P L, and Pesquet, J-C. *Proximal splitting
    methods in signal processing.* In:  Bauschke, H H, Burachik, R S,
    Combettes, P L, Elser, V, Luke, D R, and Wolkowicz, H. Fixed-point
    algorithms for inverse problems in science and engineering, Springer,
    2011.
    """

    # To begin, we could check for two things:
    # * Currently, we do not support complex scaling. We could therefore catch
    #   nonempty imaginary parts.
    # * If some components of scaling are zero, then the following routine will
    #   crash with a division-by-zero error. The correct solution would be to
    #   just keep these components and do the following computations only for
    #   the others.
    # Since these checks are computationally expensive, we do not execute them
    # unconditionally, but only if the scaling factor is a scalar:
    domain = prox_factory(1.0).domain
    if isinstance(scaling, (int, float, complex)):
        if scaling == 0:
            return proximal_const_func(domain)
        elif scaling.imag != 0:
            raise ValueError("Complex scaling not supported.")
        else:
            scaling = float(scaling.real)
    
    else:
        assert scaling in domain, f"The scaling {scaling} was passed as a {type(scaling)}, which is not supported. Please pass it either as a float or as an element of the domain of the prox_factory."

    def arg_scaling_prox_factory(sigma):
        """Create proximal for the translation with a given sigma.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``sigma * F( . * a)`` where ``sigma`` is
            the step size
        """
        scaling_square = scaling * scaling
        prox = prox_factory(sigma * scaling_square)
        space = prox.domain
        mult_inner = MultiplyOperator(scaling, domain=space, range=space)
        mult_outer = MultiplyOperator(1 / scaling, domain=space, range=space)
        return mult_outer * prox * mult_inner

    return arg_scaling_prox_factory


def proximal_quadratic_perturbation(prox_factory, a, u=None):
    r"""Calculate the proximal of function F(x) + a * \|x\|^2 + <u,x>.

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``
    a : non-negative float
        Scaling of the quadratic term
    u : Element in domain of F, optional
        Defines the linear functional. For ``None``, the zero element
        is taken.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a functional :math:`F`, this is calculated according to the rule

    .. math::
        \mathrm{prox}_{\sigma \left(F( \cdot ) + a \| \cdot \|^2 +
        <u, \cdot >\right)}(x) =
        c \; \mathrm{prox}_{\sigma F( \cdot \, c)}((x - \sigma u) c)

    where :math:`c` is the constant

    .. math::
        c = \frac{1}{\sqrt{2 \sigma a + 1}},

    :math:`a` is the scaling parameter belonging to the quadratic term,
    :math:`u` is the space element defining the linear functional, and
    :math:`\sigma` is the step size.

    For reference on the identity used, see [CP2011c]. Note that this identity
    is not the exact one given in the reference, but was recalculated for
    arbitrary step lengths.

    References
    ----------
    [CP2011c] Combettes, P L, and Pesquet, J-C. *Proximal splitting
    methods in signal processing.* In:  Bauschke, H H, Burachik, R S,
    Combettes, P L, Elser, V, Luke, D R, and Wolkowicz, H. Fixed-point
    algorithms for inverse problems in science and engineering, Springer,
    2011.
    """
    a = float(a)
    if a < 0:
        raise ValueError('scaling parameter muts be non-negative, got {}'
                         ''.format(a))

    if u is not None and not isinstance(u, LinearSpaceElement):
        raise TypeError('`u` must be `None` or a `LinearSpaceElement` '
                        'instance, got {!r}.'.format(u))

    def quadratic_perturbation_prox_factory(sigma):
        r"""Create proximal for the quadratic perturbation with a given sigma.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``sigma * (F(x) + a * \|x\|^2 + <u,x>)``,
            where ``sigma`` is the step size
        """

        const = 1.0 / math.sqrt(sigma * 2.0 * a + 1)
        prox = proximal_arg_scaling(prox_factory, const)(sigma)

        if u is not None:
            return (MultiplyOperator(const, domain=u.space, range=u.space) @
                    prox @
                    (MultiplyOperator(const, domain=u.space, range=u.space) -
                     sigma * const * u))
        else:
            space = prox.domain
            return (MultiplyOperator(const, domain=space, range=space) @
                    prox @ MultiplyOperator(const, domain=space, range=space))

    return quadratic_perturbation_prox_factory


def proximal_composition(proximal, operator, mu):
    r"""Proximal operator factory of functional composed with unitary operator.

    For a functional ``F`` and a linear unitary `Operator` ``L`` this is the
    factory for the proximal operator of ``F * L``.

    Parameters
    ----------
    proximal : callable
        A factory function that, when called with a step size returns the
        proximal operator of ``F``
    operator : `Operator`
        The operator to compose the functional with
    mu : ``operator.field`` element
        Scalar such that ``(operator.adjoint * operator)(x) = mu * x``

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a linear operator :math:`L` with the property that for a scalar
    :math:`\mu`

    .. math::
        L^*(L(x)) = \mu * x

    and a convex function :math:`F`, the following identity holds

    .. math::
        \mathrm{prox}_{\sigma F \circ L}(x) = x + \frac{1}{\mu}
        L^* \left( \mathrm{prox}_{\mu \sigma F}(Lx) - Lx \right)

    This factory function implements this functionality.

    There is no simple formula for more general operators.

    The function cannot verify that the operator is unitary, the user needs
    to verify this.

    For reference on the identity used, see [CP2011c].

    References
    ----------
    [CP2011c] Combettes, P L, and Pesquet, J-C. *Proximal splitting
    methods in signal processing.* In:  Bauschke, H H, Burachik, R S,
    Combettes, P L, Elser, V, Luke, D R, and Wolkowicz, H. Fixed-point
    algorithms for inverse problems in science and engineering, Springer,
    2011.
    """
    def proximal_composition_factory(sigma):
        """Create proximal for the dual with a given sigma

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``prox[sigma * F * L](x)``
        """
        Id = IdentityOperator(operator.domain)
        Ir = IdentityOperator(operator.range)
        prox_muf = proximal(mu * sigma)
        return (Id +
                (1.0 / mu) * operator.adjoint * ((prox_muf - Ir) * operator))

    return proximal_composition_factory


def proximal_const_func(space):
    r"""Proximal operator factory of the constant functional.

    Function to initialize the proximal operator of the constant functional
    defined on ``space``.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of the functional G=constant

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The constant functional :math:`G` is defind as :math:`G(x) = constant`
    for all values of :math:`x`. The proximal operator of this functional is
    the identity operator

    .. math::
        \mathrm{prox}_{\sigma G}(x) = x

    Note that it is independent of :math:`\sigma`.
    """
    def identity_factory(sigma):
        """Return an instance of the proximal operator.

        Parameters
        ----------
        sigma : positive float
            Unused step size parameter. Introduced to provide a unified
            interface.

        Returns
        -------
        id : `IdentityOperator`
            The proximal operator instance of G = 0 which is the
            identity operator
        """
        return IdentityOperator(space)

    return identity_factory


def proximal_box_constraint(space, lower=None, upper=None):
    r"""Proximal operator factory for ``G(x) = ind(a <= x <= b)``.

    If P is the set of elements with a <= x <= b, the indicator function of
    which is defined as::

        ind(a <= x <= b) = {0 if x in P, infinity if x is not in P}

    with x being an element in ``space``.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of the functional G(x)
    lower : ``space.field`` element or ``space`` `element-like`, optional
        The lower bound.
        Default: ``None``, interpreted as -infinity
    upper : ``space.field`` element or ``space`` `element-like`, optional
        The upper bound.
        Default: ``None``, interpreted as +infinity

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    If :math:`P` is an interval :math:`[a,b]`, the indicator function is
    defined as

    .. math::
        I_{P}(x) = \begin{cases}
        0 & \text{if } x \in P, \\
        \infty & \text{if } x \not \in P
        \end{cases}

    For a step size :math:`\sigma`, the proximal operator of
    :math:`\sigma I_{P}` is given by the projection onto the interval

    .. math::
         \mathrm{prox}_{\sigma I_{P}}(x) = \begin{cases}
         a & \text{if } x < a, \\
         x & \text{if } x \in [a,b], \\
         b & \text{if } x > b.
         \end{cases}

    The proximal operator is independent of :math:`\sigma` and invariant under
    a positive rescaling of :math:`I_{P}(x)`, since that leaves the indicator
    function unchanged.

    For spaces of the form :math:`R^n`, the definition extends naturally
    in each component.

    See Also
    --------
    proximal_nonnegativity : Special case with ``lower=0, upper=infty``
    """
    # Convert element-likes if needed, also does some space checking
    if lower is not None and lower not in space and lower not in space.field:
        lower = space.element(lower)
    if upper is not None and upper not in space and upper not in space.field:
        upper = space.element(upper)

    if lower in space.field and upper in space.field:
        if lower > upper:
            raise ValueError('invalid values, `lower` ({}) > `upper` ({})'
                             ''.format(lower, upper))

    class ProxOpBoxConstraint(Operator):

        """Proximal operator for G(x) = ind(a <= x <= b)."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter, not used.
            """
            super(ProxOpBoxConstraint, self).__init__(
                domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Apply the operator to ``x`` and store the result in ``out``."""
            if lower is not None and upper is None:
                maximum(x, lower, out=out)
            elif lower is None and upper is not None:
                minimum(x, upper, out=out)
            elif lower is not None and upper is not None:
                maximum(x, lower, out=out)
                minimum(out, upper, out=out)
            else:
                out.assign(x)

    return ProxOpBoxConstraint


def proximal_nonnegativity(space):
    """Function to create the proximal operator of ``G(x) = ind(x >= 0)``.

    Function for the proximal operator of the functional ``G(x)=ind(x >= 0)``
    to be initialized.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of the functional G(x)

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    See Also
    --------
    proximal_box_constraint
    """
    return proximal_box_constraint(space, lower=0)


def proximal_convex_conj_l2(space, lam=1, g=None):
    r"""Proximal operator factory of the convex conj of the l2-norm/distance.

    Function for the proximal operator of the convex conjugate of the
    functional F where F is the l2-norm (or distance to g, if given)::

        F(x) =  lam ||x - g||_2

    with x and g elements in ``space``, scaling factor lam, and given data g.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        An element in ``space``. Default: ``space.zero``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Most problems are forumlated for the squared norm/distance, in that case
    use the `proximal_convex_conj_l2_squared` instead.

    The :math:`L_2`-norm/distance :math:`F` is given by is given by

    .. math::
        F(x) = \lambda \|x - g\|_2

    The convex conjugate :math:`F^*` of :math:`F` is given by

    .. math::
        F^*(y) = \begin{cases}
        0 & \text{if } \|y-g\|_2 \leq \lambda, \\
        \infty & \text{else.}
        \end{cases}

    For a step size :math:`\sigma`, the proximal operator of
    :math:`\sigma F^*` is given by the projection onto the set of :math:`y`
    satisfying :math:`\|y-g\|_2 \leq \lambda`, i.e., by

    .. math::
        \mathrm{prox}_{\sigma F^*}(y) = \begin{cases}
        \lambda \frac{y - g}{\|y - g\|}
        & \text{if } \|y-g\|_2 > \lambda, \\
        y & \text{if } \|y-g\|_2 \leq \lambda
        \end{cases}

    Note that the expression is independent of :math:`\sigma`.

    See Also
    --------
    proximal_l2 : proximal without convex conjugate
    proximal_convex_conj_l2_squared : proximal for squared norm/distance
    """
    prox_l2 = proximal_l2(space, lam=lam, g=g)
    return proximal_convex_conj(prox_l2)


def proximal_l2(space, lam=1, g=None):
    r"""Proximal operator factory of the l2-norm/distance.

    Function for the proximal operator of the functional ``F`` where ``F``
    is the l2-norm (or distance to g, if given)::

        ``F(x) =  lam ||x - g||_2``

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        An element in ``space``. Default: ``space.zero``.

    Returns
    -------
    prox_factory : callable
        Factory for the proximal operator to be initialized.

    Notes
    -----
    Most problems are forumlated for the squared norm/distance, in that case
    use `proximal_l2_squared` instead.

    The :math:`L_2`-norm/distance :math:`F` is given by

    .. math::
        F(x) = \lambda \|x - g\|_2

    For a step size :math:`\sigma`, the proximal operator of :math:`\sigma F`
    is given by

    .. math::
        \mathrm{prox}_{\sigma F}(y) = \begin{cases}
        \frac{1 - c}{\|y-g\|} \cdot y  + c \cdot g
        & \text{if } c < g, \\
        g & \text{else},
        \end{cases}

    where :math:`c = \sigma \frac{\lambda}{\|y - g\|_2}`.

    See Also
    --------
    proximal_l2_squared : proximal for squared norm/distance
    proximal_convex_conj_l2 : proximal for convex conjugate
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalL2(Operator):

        """Proximal operator of the l2-norm/distance."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter
            """
            super(ProximalL2, self).__init__(
                domain=space, range=space, linear=False)
            self.sigma = float(sigma)

        def _call(self, x, out):
            """Apply the operator to ``x`` and stores the result in ``out``."""

            eps = _numerical_epsilon(self.domain)

            if g is None:
                x_norm = x.norm() * (1 + eps)
                if x_norm > 0:
                    step = self.sigma * lam / x_norm
                else:
                    step = np.inf
                
                if step < 1.0:
                    out.lincomb(1.0 - step, x)
                else:
                    out.set_zero()

            else:
                x_norm = (x - g).norm() * (1 + eps)
                if x_norm > 0:
                    step = self.sigma * lam / x_norm
                else:
                    step = np.inf

                if step < 1.0:
                    out.lincomb(1.0 - step, x, step, g)
                else:
                    out.assign(g)

    return ProximalL2


def proximal_convex_conj_l2_squared(space, lam=1, g=None):
    r"""Proximal operator factory of the convex conj of the squared l2-dist

    Function for the proximal operator of the convex conjugate of the
    functional F where F is the l2-norm (or distance to g, if given)::

        F(x) =  lam ||x - g||_2^2

    with x and g elements in ``space``, scaling factor lam, and given data g.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        An element in ``space``. Default: ``space.zero``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The squared :math:`L_2`-norm/distance :math:`F` is given by

    .. math::
        F(x) =  \lambda \|x - g\|_2^2.

    The convex conjugate :math:`F^*` of :math:`F` is given by

    .. math::
        F^*(y) = \frac{1}{4\lambda} \left( \|
        y\|_2^2 + \langle y, g \rangle \right)

    For a step size :math:`\sigma`, the proximal operator of
    :math:`\sigma F^*` is given by

    .. math::
        \mathrm{prox}_{\sigma F^*}(y) = \frac{y - \sigma g}{1 +
        \sigma/(2 \lambda)}

    See Also
    --------
    proximal_convex_conj_l2 : proximal without square
    proximal_l2_squared : proximal without convex conjugate
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalConvexConjL2Squared(Operator):

        """Proximal operator of the convex conj of the squared l2-norm/dist."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float or pointwise positive space.element
                Step size parameter. If scalar, it contains a global stepsize,
                otherwise the space.element defines a stepsize for each point.
            """
            super(ProximalConvexConjL2Squared, self).__init__(
                domain=space, range=space, linear=g is None)
            if np.isscalar(sigma):
                self.sigma = float(sigma)
            else:
                self.sigma = space.element(sigma)

        def _call(self, x, out):
            """Apply the operator to ``x`` and store the result in ``out``"""
            # (x - sig*g) / (1 + sig/(2 lam))
            sig = self.sigma
            if np.isscalar(sig):
                if g is None:
                    out.lincomb(1 / (1 + 0.5 * sig / lam), x)
                else:
                    out.lincomb(1 / (1 + 0.5 * sig / lam), x,
                                -sig / (1 + 0.5 * sig / lam), g)
            elif sig in space:
                if g is None:
                    x.divide(1 + 0.5 / lam * sig, out=out)
                else:
                    if x is out:
                        # Can't write to `out` since old `x` is still needed
                        tmp = sig.multiply(g)
                        out.lincomb(1, x, -1, tmp)
                    else:
                        sig.multiply(g, out=out)
                        out.lincomb(1, x, -1, out)
                    out.divide(1 + 0.5 / lam * sig, out=out)
            else:
                raise RuntimeError(
                    '`sigma` is neither a scalar nor a space element.'
                )

    return ProximalConvexConjL2Squared


def proximal_l2_squared(space, lam=1, g=None):
    r"""Proximal operator factory of the squared l2-norm/distance.

    Function for the proximal operator of the convex conjugate of the
    functional F where F is the l2-norm (or distance to g, if given)::

        F(x) =  lam ||x - g||_2^2

    with x and g elements in ``space``, scaling factor lam, and given data g.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        An element in ``space``. Default: ``space.zero``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The squared :math:`L_2`-norm/distance :math:`F` is given by

    .. math::
        F(x) =  \lambda \|x - g\|_2^2.

    For a step size :math:`\sigma`, the proximal operator of :math:`\sigma F`
    is given by

    .. math::
        \mathrm{prox}_{\sigma F}(x) = \frac{x + 2 \sigma \lambda g}
        {1 + 2 \sigma \lambda}.

    See Also
    --------
    proximal_l2 : proximal without square
    proximal_convex_conj_l2_squared : proximal for convex conjugate
    """
    class ProximalL2Squared(Operator):

        """Proximal operator of the squared l2-norm/dist."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float or pointwise positive space.element
                Step size parameter. If scalar, it contains a global stepsize,
                otherwise the space.element defines a stepsize for each point.
            """
            super(ProximalL2Squared, self).__init__(
                domain=space, range=space, linear=g is None)
            if np.isscalar(sigma):
                self.sigma = float(sigma)
            else:
                self.sigma = space.element(sigma)

        def _call(self, x, out):
            """Apply the operator to ``x`` and store the result in ``out``"""
            # (x + 2*sig*lam*g) / (1 + 2*sig*lam))
            sig = self.sigma
            if np.isscalar(sig):
                if g is None:
                    out.lincomb(1 / (1 + 2 * sig * lam), x)
                else:
                    out.lincomb(1 / (1 + 2 * sig * lam), x,
                                2 * sig * lam / (1 + 2 * sig * lam), g)
            else:   # sig in space
                if g is None:
                    divide(x, 1 + 2 * sig * lam, out=out)
                else:
                    if x is out:
                        # Can't write to `out` since old `x` is still needed
                        tmp = sig.multiply(2 * lam * g)
                        out.lincomb(1, x, 1, tmp)
                    else:
                        sig.multiply(2 * lam * g, out=out)
                        out.lincomb(1, x, 1, out)
                    out.divide(1 + 2 * sig * lam, out=out)

    return ProximalL2Squared


def proximal_convex_conj_l1(space, lam=1, g=None):
    r"""Proximal operator factory of the L1 norm/distance convex conjugate.

    Implements the proximal operator of the convex conjugate of the
    functional ::

        F(x) = lam ||x - g||_1

    with ``x`` and ``g`` elements in ``space``, and scaling factor ``lam``.

    Parameters
    ----------
    space : `LinearSpace` or `ProductSpace` of `LinearSpace` spaces
        Domain of the functional F
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        Element to which the L1 distance is taken.
        Default: ``space.zero``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.

    Notes
    -----
    The convex conjugate :math:`F^*` of the functional

    .. math::
        F(x) = \lambda \|x - g\|_1.

    is in the case of scalar-valued functions given by

    .. math::
        F^*(y) = \iota_{B_\infty} \big( \lambda^{-1}\, y \big) +
        \left\langle \lambda^{-1}\, y,\: g \right\rangle,

    where :math:`\iota_{B_\infty}` is the indicator function of the
    unit ball with respect to :math:`\|\cdot\|_\infty`.
    For vector-valued functions, the convex conjugate is

    .. math::
        F^*(y) = \sum_{k=1}^d F^*(y_k)

    due to separability of the (non-isotropic) 1-norm.

    For a step size :math:`\sigma`, the proximal operator of
    :math:`\sigma F^*` is given by

    .. math::
        \mathrm{prox}_{\sigma F^*}(y) = \frac{\lambda (y - \sigma g)}{
        \max(\lambda, |y - \sigma g|)}

    Here, all operations are to be read pointwise.

    For vector-valued :math:`x` and :math:`g`, the (non-isotropic) proximal
    operator is the component-wise scalar proximal:

    .. math::
        \mathrm{prox}_{\sigma F^*}(x) = \left(
            \mathrm{prox}_{\sigma F^*}(x_1), \dots,
            \mathrm{prox}_{\sigma F^*}(x_d)
            \right),

    where :math:`d` is the number of components of :math:`x`.

    See Also
    --------
    proximal_convex_conj_l1_l2 : isotropic variant for vector-valued functions
    proximal_l1 : proximal without convex conjugate
    """
    # Fix for rounding errors
    eps = _numerical_epsilon(space)
    lam = float(lam * (1 - eps))

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalConvexConjL1(Operator):

        """Proximal operator of the L1 norm/distance convex conjugate."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float or pointwise positive space.element
                Step size parameter. If scalar, it contains a global stepsize,
                otherwise the space.element defines a stepsize for each point.
            """
            super(ProximalConvexConjL1, self).__init__(
                domain=space, range=space, linear=False)
            if np.isscalar(sigma):
                self.sigma = float(sigma)
            else:
                self.sigma = space.element(sigma)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            # lam * (x - sig * g) / max(lam, |x - sig * g|)

            # diff = x - sig * g
            if g is not None:
                diff = self.domain.element()
                diff.lincomb(1, x, -self.sigma, g)
            else:
                if x is out:
                    # Handle aliased `x` and `out`
                    # This is necessary since we write to both `diff` and
                    # `out`.
                    diff = x.copy()
                else:
                    diff = x

            # out = max( |x-sig*g|, lam ) / lam
            abs(diff, out=out)
            maximum(out, lam, out=out)
            out /= lam

            # out = diff / ...
            divide(diff, out, out=out)

    return ProximalConvexConjL1


def proximal_convex_conj_l1_l2(space, lam=1, g=None):
    r"""Proximal operator factory of the L1-L2 norm/distance convex conjugate.

    Implements the proximal operator of the convex conjugate of the
    functional ::

        F(x) = lam || |x - g|_2 ||_1

    with ``x`` and ``g`` elements in ``space``, and scaling factor ``lam``.
    Here, ``|.|_2`` is the pointwise Euclidean norm of a vector-valued
    function.

    Parameters
    ----------
    space : `LinearSpace` or `ProductSpace` of `LinearSpace` spaces
        Domain of the functional F
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        Element to which the L1 distance is taken.
        Default: ``space.zero``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.

    Notes
    -----
    The convex conjugate :math:`F^*` of the functional

    .. math::
        F(x) = \lambda \| |x - g|_2 \|_1.

    is given by

    .. math::
        F^*(y) = \iota_{B_\infty} \big( \lambda^{-1}\, |y|_2 \big) +
        \left\langle \lambda^{-1}\, y,\: g \right\rangle,

    where :math:`\iota_{B_\infty}` is the indicator function of the
    unit ball with respect to :math:`\|\cdot\|_\infty`.

    For a step size :math:`\sigma`, the proximal operator of
    :math:`\sigma F^*` is given by

    .. math::
        \mathrm{prox}_{\sigma F^*}(y) = \frac{\lambda (y - \sigma g)}{
        \max(\lambda, |y - \sigma g|_2)}

    Here, all operations are to be read pointwise.

    See Also
    --------
    proximal_convex_conj_l1 : Scalar or non-isotropic vectorial variant
    """
    # Fix for rounding errors
    eps = _numerical_epsilon(space)
    lam = float(lam * (1 - eps))

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalConvexConjL1L2(Operator):

        """Proximal operator of the convex conj of the l1-norm/distance."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter
            """
            super(ProximalConvexConjL1L2, self).__init__(
                domain=space, range=space, linear=False)
            self.sigma = float(sigma)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            # lam * (x - sig * g) / max(lam, |x - sig * g|)

            # diff = x - sig * g
            if g is not None:
                diff = self.domain.element()
                diff.lincomb(1, x, -self.sigma, g)
            else:
                diff = x

            # denom = max( |x-sig*g|_2, lam ) / lam  (|.|_2 pointwise)
            pwnorm = PointwiseNorm(self.domain, exponent=2)
            denom = pwnorm(diff)
            maximum(denom, lam, out=denom)
            denom /= lam

            # Pointwise division
            for out_i, diff_i in zip(out, diff):
                diff_i.divide(denom, out=out_i)

    return ProximalConvexConjL1L2


def proximal_l1(space, lam=1, g=None):
    r"""Proximal operator factory of the L1 norm/distance.

    Implements the proximal operator of the functional ::

        F(x) = lam ||x - g||_1

    with ``x`` and ``g`` elements in ``space``, and scaling factor ``lam``.

    Parameters
    ----------
    space : `LinearSpace` or `ProductSpace`
        Domain of the functional.
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        Element to which the L1 distance is taken.
        Default: ``space.zero``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    For the functional

    .. math::
        F(x) = \lambda \|x - g\|_1,

    and a step size :math:`\sigma`, the proximal operator of
    :math:`\sigma F` is given as the "soft-shrinkage" operator

    .. math::
        \mathrm{prox}_{\sigma F}(x) =
        \begin{cases}
            g, & \text{where } |x - g| \leq \sigma\lambda, \\
            x - \sigma\lambda \mathrm{sign}(x - g), & \text{elsewhere.}
        \end{cases}

    Here, all operations are to be read pointwise.

    For vector-valued :math:`x` and :math:`g`, the (non-isotropic) proximal
    operator is the component-wise scalar proximal:

    .. math::
        \mathrm{prox}_{\sigma F}(x) = \left(
            \mathrm{prox}_{\sigma F}(x_1), \dots,
            \mathrm{prox}_{\sigma F}(x_d)
            \right),

    where :math:`d` is the number of components of :math:`x`.

    See Also
    --------
    proximal_convex_conj_l1 : proximal for convex conjugate
    proximal_l1_l2 : isotropic variant of the group L1 norm proximal
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalL1(Operator):

        """Proximal operator of the L1 norm/distance."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float or pointwise positive space.element
                Step size parameter. If scalar, it contains a global stepsize,
                otherwise the space.element defines a stepsize for each point.
            """
            super(ProximalL1, self).__init__(
                domain=space, range=space, linear=False)
            if np.isscalar(sigma):
                self.sigma = float(sigma)
            else:
                self.sigma = space.element(sigma)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            # diff = x - g
            if g is not None:
                diff = x - g
            else:
                if x is out:
                    # Handle aliased `x` and `out` (original `x` needed later)
                    diff = x.copy()
                else:
                    diff = x

            # We write the operator as
            # x - (x - g) / max(|x - g| / sig*lam, 1)
            denom = abs(diff)
            denom /= self.sigma * lam
            maximum(denom, 1, out=denom)

            # out = (x - g) / denom
            divide(diff, denom, out=out)

            # out = x - ...
            out.lincomb(1, x, -1, out)

    return ProximalL1


def proximal_l1_l2(space, lam=1, g=None):
    r"""Proximal operator factory of the group-L1-L2 norm/distance.

    Implements the proximal operator of the functional ::

        F(x) = lam || |x - g|_2 ||_1

    with ``x`` and ``g`` elements in ``space``, and scaling factor ``lam``.
    Here, ``|.|_2`` is the pointwise Euclidean norm of a vector-valued
    function.

    Parameters
    ----------
    space : `LinearSpace` or `ProductSpace`
        Domain of the functional.
    lam : positive float, optional
        Scaling factor or regularization parameter.
    g : ``space`` element, optional
        Element to which the L1-L2 distance is taken.
        Default: ``space.zero``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    For the functional

    .. math::
        F(x) = \lambda \| |x - g|_2 \|_1,

    and a step size :math:`\sigma`, the proximal operator of
    :math:`\sigma F` is given as the "soft-shrinkage" operator

    .. math::
        \mathrm{prox}_{\sigma F}(x) =
        \begin{cases}
            g, & \text{where } |x - g|_2 \leq \sigma\lambda, \\
            x - \sigma\lambda \frac{x - g}{|x - g|_2}, & \text{elsewhere.}
        \end{cases}

    Here, all operations are to be read pointwise.

    See Also
    --------
    proximal_l1 : Scalar or non-isotropic vectorial variant
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalL1L2(Operator):

        """Proximal operator of the group-L1-L2 norm/distance."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter.
            """
            super(ProximalL1L2, self).__init__(
                domain=space, range=space, linear=False)
            self.sigma = float(sigma)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            # diff = x - g
            if g is not None:
                diff = x - g
            else:
                if x is out:
                    # Handle aliased `x` and `out` (original `x` needed later)
                    diff = x.copy()
                else:
                    diff = x

            # We write the operator as
            # x - (x - g) / max(|x - g|_2 / sig*lam, 1)
            pwnorm = PointwiseNorm(self.domain, exponent=2)
            denom = pwnorm(diff)
            denom /= self.sigma * lam
            maximum(denom, 1, out=denom)

            # out = (x - g) / denom
            for out_i, diff_i in zip(out, diff):
                divide(diff_i, denom, out=out_i)

            # out = x - ...
            out.lincomb(1, x, -1, out)

    return ProximalL1L2


def proximal_linfty(space):
    r"""Proximal operator factory of the ``l_\infty``-norm.

    Function for the proximal operator of the functional ``F`` where ``F``
    is the ``l_\infty``-norm::

        ``F(x) =  \sup_i |x_i|``

    Parameters
    ----------
    space : `LinearSpace`
        Domain of ``F``.

    Returns
    -------
    prox_factory : callable
        Factory for the proximal operator to be initialized.

    Notes
    -----
    The proximal is computed by the Moreau identity and a projection onto an
    l1-ball [PB2014].

    See Also
    --------
    proj_l1 : projection onto l1-ball
    """
    class ProximalLInfty(Operator):

        """Proximal operator of the linf-norm."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter
            """
            super(ProximalLInfty, self).__init__(
                domain=space, range=space, linear=False)
            self.sigma = float(sigma)

        def _call(self, x, out):
            """Return ``self(x)``."""

            radius = self.sigma

            if x is out:
                x = x.copy()

            proj_l1(x, radius, out)
            out.lincomb(-1, out, 1, x)

    return ProximalLInfty


def proximal_convex_conj_linfty(space):
    r"""Proximal operator factory of the Linfty norm/distance convex conjugate.

    Implements the proximal operator of the convex conjugate of the
    functional ::

        F(x) = \|x\|_\infty

    with ``x`` in ``space``.

    Parameters
    ----------
    space : `LinearSpace` or `ProductSpace` of `LinearSpace` spaces
        Domain of the functional F

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.

    Notes
    -----
    The convex conjugate :math:`F^*` of the functional

    .. math::
        F(x) = \|x\|_\infty.

    is in the case of scalar-valued functions given by the indicator function
    of the unit 1-norm ball

    .. math::
        F^*(y) = \iota_{B_1} \big( y \big).

    See Also
    --------
    proj_l1 : orthogonal projection onto balls in the 1-norm
    """

    class ProximalConvexConjLinfty(Operator):

        """Proximal operator of the Linfty norm/distance convex conjugate."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float or pointwise positive space.element
                Step size parameter. If scalar, it contains a global stepsize,
                otherwise the space.element defines a stepsize for each point.
            """
            super(ProximalConvexConjLinfty, self).__init__(
                domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            proj_l1(x, radius=1, out=out)

    return ProximalConvexConjLinfty


def proj_l1(x, radius=1, out=None):
    r"""Projection onto l1-ball.

    Projection onto::

        ``{ x \in X | ||x||_1 \leq r}``

    with ``r`` being the radius.

    Parameters
    ----------
    space : `LinearSpace`
        Space / domain ``X``.
    radius : positive float, optional
        Radius ``r`` of the ball.

    Returns
    -------
    prox_factory : callable
        Factory for the proximal operator to be initialized.

    Notes
    -----
    The projection onto an l1-ball can be computed by projection onto a
    simplex, see [D+2008] for details.

    References
    ----------
    [D+2008] Duchi, J., Shalev-Shwartz, S., Singer, Y., and Chandra, T.
    *Efficient Projections onto the L1-ball for Learning in High dimensions*.
    ICML 2008, pp. 272-279. http://doi.org/10.1145/1390156.1390191

    See Also
    --------
    proximal_linfty : proximal for l-infinity norm
    proj_simplex : projection onto simplex
    """

    if out is None:
        out = x.space.element()

    u = abs(x)
    if sum(u) <= radius:
        out[:] = x
    else:
        v = sign(u)
        proj_simplex(u, radius, out)
        out *= v

    return out


def proj_simplex(x, diameter=1, out=None):
    r"""Projection onto simplex.

    Projection onto::

        ``{ x \in X | x_i \geq 0, \sum_i x_i = r}``

    with :math:`r` being the diameter. It is computed by the formula proposed
    in [D+2008].

    Parameters
    ----------
    space : `LinearSpace`
        Space / domain ``X``.
    diameter : positive float, optional
        Diameter of the simplex.

    Returns
    -------
    prox_factory : callable
        Factory for the proximal operator to be initialized.

    Notes
    -----
    The projection onto a simplex is not of closed-form but can be solved by a
    non-iterative algorithm, see [D+2008] for details.

    References
    ----------
    [D+2008] Duchi, J., Shalev-Shwartz, S., Singer, Y., and Chandra, T.
    *Efficient Projections onto the L1-ball for Learning in High dimensions*.
    ICML 2008, pp. 272-279. http://doi.org/10.1145/1390156.1390191

    See Also
    --------
    proj_l1 : projection onto l1-norm ball
    """
    if out is None:
        out = x.space.element()

    # sort values in descending order
    x_sor = x.asarray().flatten()
    x_sor.sort()
    x_sor = x_sor[::-1]

    # find critical index
    j = np.arange(1, x.size + 1)
    x_avrg = (1 / j) * (np.cumsum(x_sor) - diameter)
    crit = x_sor - x_avrg
    i = np.argwhere(crit >= 0).flatten().max()

    # output is a shifted and thresholded version of the input
    out[:] = np.maximum(x - x_avrg[i], 0)

    return out


def proximal_convex_conj_kl(space, lam=1, g=None):
    r"""Proximal operator factory of the convex conjugate of the KL divergence.

    Function returning the proximal operator of the convex conjugate of the
    functional F where F is the entropy-type Kullback-Leibler (KL) divergence::

        F(x) = sum_i (x_i - g_i + g_i ln(g_i) - g_i ln(pos(x_i))) + ind_P(x)

    with ``x`` and ``g`` elements in the linear space ``X``, and ``g``
    non-negative. Here, ``pos`` denotes the nonnegative part, and ``ind_P`` is
    the indicator function for nonnegativity.

    Parameters
    ----------
    space : `TensorSpace`
        Space X which is the domain of the functional F
    lam : positive float, optional
        Scaling factor.
    g : ``space`` element, optional
        Data term, positive. If None it is take as the one-element.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.

    See Also
    --------
    proximal_convex_conj_kl_cross_entropy : proximal for releated functional

    Notes
    -----
    The functional is given by the expression

    .. math::
        F(x) = \sum_i (x_i - g_i + g_i \ln(g_i) - g_i \ln(pos(x_i))) +
        I_{x \geq 0}(x)

    The indicator function :math:`I_{x \geq 0}(x)` is used to restrict the
    domain of :math:`F` such that :math:`F` is defined over whole space
    :math:`X`. The non-negativity thresholding :math:`pos` is used to define
    :math:`F` in the real numbers.

    Note that the functional is not well-defined without a prior g. Hence, if g
    is omitted this will be interpreted as if g is equal to the one-element.

    The convex conjugate :math:`F^*` of :math:`F` is

    .. math::
        F^*(p) = \sum_i (-g_i \ln(\text{pos}({1_X}_i - p_i))) +
        I_{1_X - p \geq 0}(p)

    where :math:`p` is the variable dual to :math:`x`, and :math:`1_X` is an
    element of the space :math:`X` with all components set to 1.

    The proximal operator of the convex conjugate of F is

    .. math::
        \mathrm{prox}_{\sigma (\lambda F)^*}(x) =
        \frac{\lambda 1_X + x - \sqrt{(x -  \lambda 1_X)^2 +
        4 \lambda \sigma g}}{2}

    where :math:`\sigma` is the step size-like parameter, and :math:`\lambda`
    is the weighting in front of the function :math:`F`.

    KL based objectives are common in MLEM optimization problems and are often
    used when data noise governed by a multivariate Poisson probability
    distribution is significant.

    The intermediate image estimates can have negative values even though
    the converged solution will be non-negative. Non-negative intermediate
    image estimates can be enforced by adding an indicator function ind_P
    the primal objective.

    This functional :math:`F`, described above, is related to the
    Kullback-Leibler cross entropy functional. The KL cross entropy is the one
    described in `this Wikipedia article
    <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence>`_, and
    the functional :math:`F` is obtained by switching place of the prior and
    the varialbe in the KL cross entropy functional. See the See Also section.
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{} is not an element of {}'.format(g, space))

    class ProximalConvexConjKL(Operator):

        """Proximal operator of the convex conjugate of the KL divergence."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
            """
            super(ProximalConvexConjKL, self).__init__(
                domain=space, range=space, linear=False)
            self.sigma = float(sigma)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            # (x + lam - sqrt((x - lam)^2 + 4*lam*sig*g)) / 2

            # out = (x - lam)^2
            if x is out:
                # Handle aliased `x` and `out` (need original `x` later on)
                x = x.copy()
            else:
                out.assign(x)
            out -= lam
            square(out, out=out)

            # out = ... + 4*lam*sigma*g
            # If g is None, it is taken as the one element
            if g is None:
                out += 4.0 * lam * self.sigma
            else:
                out.lincomb(1, out, 4.0 * lam * self.sigma, g)

            # out = x - sqrt(...) + lam
            sqrt(out, out=out)
            out.lincomb(1, x, -1, out)
            out += lam

            # out = 1/2 * ...
            out /= 2

    return ProximalConvexConjKL


def proximal_convex_conj_kl_cross_entropy(space, lam=1, g=None):
    r"""Proximal factory of the convex conj of cross entropy KL divergence.

    Function returning the proximal factory of the convex conjugate of the
    functional F, where F is the cross entropy Kullback-Leibler (KL)
    divergence given by::

        F(x) = sum_i (x_i ln(pos(x_i)) - x_i ln(g_i) + g_i - x_i) + ind_P(x)

    with ``x`` and ``g`` in the linear space ``X``, and ``g`` non-negative.
    Here, ``pos`` denotes the nonnegative part, and ``ind_P`` is the indicator
    function for nonnegativity.

    Parameters
    ----------
    space : `TensorSpace`
        Space X which is the domain of the functional F
    lam : positive float, optional
        Scaling factor.
    g : ``space`` element, optional
        Data term, positive. If None it is take as the one-element.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.

    See Also
    --------
    proximal_convex_conj_kl : proximal for related functional

    Notes
    -----
    The functional is given by the expression

    .. math::
        F(x) = \sum_i (x_i \ln(pos(x_i)) - x_i \ln(g_i) + g_i - x_i) +
        I_{x \geq 0}(x)

    The indicator function :math:`I_{x \geq 0}(x)` is used to restrict the
    domain of :math:`F` such that :math:`F` is defined over whole space
    :math:`X`. The non-negativity thresholding :math:`pos` is used to define
    :math:`F` in the real numbers.

    Note that the functional is not well-defined without a prior g. Hence, if g
    is omitted this will be interpreted as if g is equal to the one-element.

    The convex conjugate :math:`F^*` of :math:`F` is

    .. math::
        F^*(p) = \sum_i g_i (exp(p_i) - 1)

    where :math:`p` is the variable dual to :math:`x`.

    The proximal operator of the convex conjugate of :math:`F` is

    .. math::
        \mathrm{prox}_{\sigma (\lambda F)^*}(x) = x - \lambda
        W(\frac{\sigma}{\lambda} g e^{x/\lambda})

    where :math:`\sigma` is the step size-like parameter, :math:`\lambda` is
    the weighting in front of the function :math:`F`, and :math:`W` is the
    Lambert W function (see, for example, the
    `Wikipedia article <https://en.wikipedia.org/wiki/Lambert_W_function>`_).

    For real-valued input x, the Lambert :math:`W` function is defined only for
    :math:`x \geq -1/e`, and it has two branches for values
    :math:`-1/e \leq x < 0`. However, for inteneded use-cases, where
    :math:`\lambda` and :math:`g` are positive, the argument of :math:`W`
    will always be positive.

    `Wikipedia article on Kullback Leibler divergence
    <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence>`_.
    For further information about the functional, see for example `this article
    <http://ieeexplore.ieee.org/document/1056144/?arnumber=1056144>`_.

    The KL cross entropy functional :math:`F`, described above, is related to
    another functional functional also know as KL divergence. This functional
    is often used as data discrepancy term in inverse problems, when data is
    corrupted with Poisson noise. This functional is obtained by changing place
    of the prior and the variable. See the See Also section.
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{} is not an element of {}'.format(g, space))

    class ProximalConvexConjKLCrossEntropy(Operator):

        """Proximal operator of conjugate of cross entropy KL divergence."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
            """
            self.sigma = float(sigma)
            nonlocal g
            self.g = g
            super(ProximalConvexConjKLCrossEntropy, self).__init__(
                domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            # Lazy import to improve `import odl` time
            if isinstance(x, ProductSpaceElement) and x[0].space.device!= 'cpu':
                warnings.warn(f'The function ``_call`` of ``ProximalConvexConjKLCrossEntropy`` involves a ``lambertw`` call. At present, ODL relies on scipy to perform it and it does not support GPU inputs for that specific function. As such, the input will be moved to the cpu, which will slow down the algorithm.', stacklevel=2)
                # FML
                namespace = x[0].space.array_namespace
                if g is None:
                    lambw = [scipy_lambertw(
                        (self.sigma / lam) * namespace.exp(sub_x.to('cpu') / lam)) for sub_x in x.asarray()]
                else:
                    lambw = [scipy_lambertw(
                        (self.sigma / lam) * sub_g.to('cpu')*  namespace.exp(sub_x.to('cpu') / lam)) for (sub_g, sub_x) in zip(self.g.asarray(), x.asarray())]
                    if not is_complex_dtype(self.domain.dtype):
                        lambw = [lambw_.real for lambw_ in lambw]
            elif isinstance(x, Tensor) and x.space.device!= 'cpu':
                namespace = x.space.array_namespace
                if g is None:
                    lambw = scipy_lambertw(
                        (self.sigma / lam) * namespace.exp(x.asarray().to('cpu') / lam))
                else:
                    lambw = scipy_lambertw(
                        (self.sigma / lam) * self.g.asarray().to('cpu')*  namespace.exp(x.asarray().to('cpu') / lam)) 
                    if not is_complex_dtype(self.domain.dtype):
                        lambw = [lambw_.real for lambw_ in lambw]
            else:
                print('ELSE branch')
                print(type(x))
                if g is None:
                    # If g is None, it is taken as the one element
                    # Different branches of lambertw is not an issue, see Notes
                    lambw = lambertw(
                        (self.sigma / lam) * exp(x / lam))
                else:
                    # Different branches of lambertw is not an issue, see Notes
                    lambw = lambertw(
                        (self.sigma / lam) * self.g * exp(x / lam))

                if not is_complex_dtype(self.domain.dtype):
                    lambw = lambw.real

            lambw = x.space.element(lambw)

            out.lincomb(1, x, -lam, lambw)

    return ProximalConvexConjKLCrossEntropy


def proximal_huber(space, gamma):
    """Proximal factory of the Huber norm.

    Parameters
    ----------
    space : `TensorSpace`
        The domain of the functional
    gamma : float
        The smoothing parameter of the Huber norm functional.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.

    See Also
    --------
    odl.solvers.default_functionals.Huber : the Huber norm functional

    Notes
    -----
    The proximal operator is given by given by the proximal operator of
    ``1/(2*gamma) * L2 norm`` in points that are ``<= gamma``, and by the
    proximal operator of the l1 norm in points that are ``> gamma``.
    """

    gamma = float(gamma)

    class ProximalHuber(Operator):

        """Proximal operator of Huber norm."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
            """
            self.sigma = float(sigma)
            super(ProximalHuber, self).__init__(domain=space, range=space,
                                                linear=False)

        def _call(self, x, out):
            """Return ``self(x, out=out)``."""
            if isinstance(self.domain, ProductSpace):
                norm = PointwiseNorm(self.domain, 2)(x)
            else:
                norm = abs(x)

            mask = less_equal(norm, gamma + self.sigma)
            out[mask] = gamma / (gamma + self.sigma) * x[mask]

            logical_not(mask, out=mask)
            sign_x = sign(x)
            out[mask] = x[mask] - self.sigma * sign_x[mask]

            return out

    return ProximalHuber


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for the ODL-pyshearlab integration."""

import pytest
import numpy as np
import odl
import odl.contrib.pyshearlab
from odl.core.util.testutils import all_almost_equal, simple_fixture


dtype = simple_fixture('dtype', ['float32', 'float64'])
shape = simple_fixture('shape', [(64, 64), (128, 128)])


def test_operator(dtype, shape):
    """Test PyShearlabOperator operator."""
    rel = 100 * np.finfo('float32').resolution

    space = odl.uniform_discr([-1, -1], [1, 1], shape, dtype=dtype)

    op = odl.contrib.pyshearlab.PyShearlabOperator(space, num_scales=2)

    phantom = odl.core.phantom.shepp_logan(space, True)

    # Test evaluation
    y = op(phantom)

    # <Ax, Ax> = <x, AtAx>
    ax = op.adjoint(y)
    assert pytest.approx(y.inner(y), rel=rel) == phantom.inner(ax)

    # A^{-1} A x = x
    rec = op.inverse(y)
    assert all_almost_equal(op.inverse(y), phantom)

    # <A^{-1}y, A^{-1}y> = <y, A^{-*}A^{-1}y>
    recadj = op.inverse.adjoint(rec)
    assert pytest.approx(rec.inner(rec), rel=rel) == y.inner(recadj)

    # A^{-*}A^*y = y
    adjinvadj = op.adjoint.inverse(op.adjoint(y))
    assert all_almost_equal(adjinvadj, y, ndigits=5)

    # A^*A^{-*}x = x
    adjadjinv = op.adjoint(op.adjoint.inverse(phantom))
    assert all_almost_equal(adjadjinv, phantom, ndigits=5)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test ASTRA setup functions."""

from __future__ import division

import numpy as np
import pytest

import odl
from odl.applications.tomo.backends.astra_setup import (
    astra_algorithm, astra_data, astra_projection_geometry, astra_projector,
    astra_supports, astra_volume_geometry)
from odl.core.util.testutils import is_subdict
from odl.core.util.testutils import simple_fixture
try:
    import astra
except ImportError:
    pass

pytestmark = pytest.mark.skipif("not odl.applications.tomo.ASTRA_AVAILABLE")

astra_impl = simple_fixture('astra_impl', params=['cpu', 'cuda'])


def _discrete_domain(ndim, impl, device):
    """Create `DiscretizedSpace` space with isotropic grid stride.

    Parameters
    ----------
    ndim : `int`
        Number of space dimensions

    Returns
    -------
    space : `DiscretizedSpace`
        Returns a `DiscretizedSpace` instance
    """
    max_pt = np.arange(1, ndim + 1)
    min_pt = -max_pt
    shape = np.arange(1, ndim + 1) * 10

    return odl.uniform_discr(min_pt, max_pt, impl=impl, device=device, shape=shape, dtype='float32')


def _discrete_domain_anisotropic(ndim, impl, device):
    """Create `DiscretizedSpace` space with anisotropic grid stride.

    Parameters
    ----------
    ndim : `int`
        Number of space dimensions

    Returns
    -------
    space : `DiscretizedSpace`
        Returns a `DiscretizedSpace` instance
    """
    min_pt = [-1] * ndim
    max_pt = [1] * ndim
    shape = np.arange(1, ndim + 1) * 10

    return odl.uniform_discr(min_pt, max_pt, impl=impl, device=device, shape=shape, dtype='float32')



def test_vol_geom_2d(astra_impl, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Check correctness of ASTRA 2D volume geometries."""
    x_pts = 10  # x_pts = Rows
    y_pts = 20  # y_pts = Columns

    # Isotropic voxel case
    discr_dom = _discrete_domain(2, impl=impl, device=device)
    correct_dict = {
        'cpu': {
            'GridColCount': y_pts,
            'GridRowCount': x_pts,
            'option': {
                'WindowMinX': -2.0,  # y_min
                'WindowMaxX': 2.0,   # y_max
                'WindowMinY': -1.0,  # x_min
                'WindowMaxY': 1.0}   # x_amx
            },
        'cuda':  {
            'GridColCount': y_pts,
            'GridRowCount': x_pts,
            'GridSliceCount': 1,
            'option': {
                'WindowMinX': -2.0,  # y_min
                'WindowMaxX': 2.0,   # y_max
                'WindowMinY': -1.0,  # x_min
                'WindowMaxY': 1.0,   # x_amx
                'WindowMinZ': -1,    # z_min
                'WindowMaxZ': 1,     # z_max
                }
            }  
    }

    vol_geom = astra_volume_geometry(discr_dom, astra_impl)
    assert vol_geom == correct_dict[astra_impl]

    # Anisotropic voxel case
    discr_dom = _discrete_domain_anisotropic(2, impl=impl, device=device)
    correct_dict = {
        'cpu': {
        'GridColCount': y_pts,
        'GridRowCount': x_pts,
        'option': {
            'WindowMinX': -1.0,  # y_min
            'WindowMaxX': 1.0,  # y_max
            'WindowMinY': -1.0,  # x_min
            'WindowMaxY': 1.0}   # x_max
        },
        'cuda':  {
        'GridColCount': y_pts,
        'GridRowCount': x_pts,
        'GridSliceCount': 1,
        'option': {
            'WindowMinX': -1.0,  # y_min
            'WindowMaxX': 1.0,   # y_max
            'WindowMinY': -1.0,  # x_min
            'WindowMaxY': 1.0,   # x_amx
            'WindowMinZ': -1,    # z_min
            'WindowMaxZ': 1,     # z_max
            }
        }  
    }

    if astra_supports('anisotropic_voxels_2d'):
        vol_geom = astra_volume_geometry(discr_dom,astra_impl)
        assert vol_geom == correct_dict[astra_impl]
    else:
        with pytest.raises(NotImplementedError):
            astra_volume_geometry(discr_dom, astra_impl)


def test_vol_geom_3d(astra_impl, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Check correctness of ASTRA 3D volume geometies."""
    x_pts = 10
    y_pts = 20
    z_pts = 30

    # Isotropic voxel case
    discr_dom = _discrete_domain(3, impl=impl, device=device)
    # x = columns, y = rows, z = slices
    correct_dict = {
        'GridColCount': z_pts,
        'GridRowCount': y_pts,
        'GridSliceCount': x_pts,
        'option': {
            'WindowMinX': -3.0,  # z_min
            'WindowMaxX': 3.0,  # z_max
            'WindowMinY': -2.0,  # y_min
            'WindowMaxY': 2.0,  # y_amx
            'WindowMinZ': -1.0,  # x_min
            'WindowMaxZ': 1.0}}  # x_amx

    vol_geom = astra_volume_geometry(discr_dom, astra_impl)
    assert vol_geom == correct_dict

    discr_dom = _discrete_domain_anisotropic(3, impl=impl, device=device)
    # x = columns, y = rows, z = slices
    correct_dict = {
        'GridColCount': z_pts,
        'GridRowCount': y_pts,
        'GridSliceCount': x_pts,
        'option': {
            'WindowMinX': -1.0,  # z_min
            'WindowMaxX': 1.0,  # z_max
            'WindowMinY': -1.0,  # y_min
            'WindowMaxY': 1.0,  # y_amx
            'WindowMinZ': -1.0,  # x_min
            'WindowMaxZ': 1.0}}  # x_amx

    if astra_supports('anisotropic_voxels_3d'):
        vol_geom = astra_volume_geometry(discr_dom, astra_impl)
        assert vol_geom == correct_dict
    else:
        with pytest.raises(NotImplementedError):
            astra_volume_geometry(discr_dom, astra_impl)


def test_proj_geom_parallel_2d(astra_impl):
    """Create ASTRA 2D projection geometry."""

    apart = odl.uniform_partition(0, 2, 5)
    dpart = odl.uniform_partition(-1, 1, 10)
    geom = odl.applications.tomo.Parallel2dGeometry(apart, dpart)

    proj_geom = astra_projection_geometry(geom, astra_impl)

    correct_subdict = {
        'cpu' : {
            'type': 'parallel',
            'DetectorCount': 10, 
            'DetectorWidth': 0.2
            },
        'cuda' : {
            'type': 'parallel3d_vec',
            'DetectorRowCount': 1, 
            'DetectorColCount': 10
            },
    }

    assert is_subdict(correct_subdict[astra_impl], proj_geom)
    if astra_impl == 'cpu':
        assert 'ProjectionAngles' in proj_geom
    else:
        assert 'Vectors' in proj_geom


def test_astra_projection_geometry(astra_impl):
    """Create ASTRA projection geometry from geometry objects."""

    with pytest.raises(TypeError):
        astra_projection_geometry(None, astra_impl=astra_impl)

    apart = odl.uniform_partition(0, 2 * np.pi, 5)
    dpart = odl.uniform_partition(-40, 40, 10)

    # motion sampling grid, detector sampling grid but not uniform
    dpart_0 = odl.RectPartition(odl.IntervalProd(0, 3),
                                odl.RectGrid([0, 1, 3]))
    geom_p2d = odl.applications.tomo.Parallel2dGeometry(apart, dpart=dpart_0)
    with pytest.raises(ValueError):
        astra_projection_geometry(geom_p2d, astra_impl)

    # detector sampling grid, motion sampling grid
    geom_p2d = odl.applications.tomo.Parallel2dGeometry(apart, dpart)
    astra_projection_geometry(geom_p2d, astra_impl)

    # Parallel 2D geometry
    geom_p2d = odl.applications.tomo.Parallel2dGeometry(apart, dpart)
    astra_geom = astra_projection_geometry(geom_p2d, astra_impl)
    if astra_impl == 'cpu':
        assert astra_geom['type'] == 'parallel'
    else:
        assert astra_geom['type'] == 'parallel3d_vec'
    # Fan flat
    src_rad = 10
    det_rad = 5
    geom_ff = odl.applications.tomo.FanBeamGeometry(apart, dpart, src_rad, det_rad)
    astra_geom = astra_projection_geometry(geom_ff, astra_impl)
    if astra_impl == 'cpu':
        assert astra_geom['type'] == 'fanflat_vec'
    else:
        assert astra_geom['type'] == 'cone_vec'

    dpart = odl.uniform_partition([-40, -3], [40, 3], (10, 5))

    # Parallel 3D geometry
    geom_p3d = odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart)
    astra_projection_geometry(geom_p3d,astra_impl)
    astra_geom = astra_projection_geometry(geom_p3d, astra_impl)
    assert astra_geom['type'] == 'parallel3d_vec'

    # Circular conebeam flat
    geom_ccf = odl.applications.tomo.ConeBeamGeometry(apart, dpart, src_rad, det_rad)
    astra_geom = astra_projection_geometry(geom_ccf, astra_impl)
    assert astra_geom['type'] == 'cone_vec'

    # Helical conebeam flat
    pitch = 1
    geom_hcf = odl.applications.tomo.ConeBeamGeometry(apart, dpart, src_rad, det_rad,
                                         pitch=pitch)
    astra_geom = astra_projection_geometry(geom_hcf, astra_impl)
    assert astra_geom['type'] == 'cone_vec'


VOL_GEOM_2D = {
    'GridColCount': 20, 'GridRowCount': 10,
    'option': {'WindowMinX': -2.0, 'WindowMaxX': 2.0,
               'WindowMinY': -1.0, 'WindowMaxY': 1.0}}


def test_volume_data_2d():
    """Create ASTRA data structure in 2D."""
    # From scratch
    data_id = astra_data(VOL_GEOM_2D, 'volume', ndim=2)
    data_out = astra.data2d.get_shared(data_id)
    assert data_out.shape == (10, 20)

    # From existing
    discr_dom = _discrete_domain(2, impl='numpy', device='cpu')
    data_in = discr_dom.element(np.ones((10, 20), dtype='float32'))
    data_id = astra_data(VOL_GEOM_2D, 'volume', data=data_in)
    data_out = astra.data2d.get_shared(data_id)
    assert data_out.shape == (10, 20)


VOL_GEOM_3D = {
    'GridColCount': 30, 'GridRowCount': 20, 'GridSliceCount': 10,
    'option': {}}


def test_volume_data_3d():
    """Create ASTRA data structure in 2D."""

    # From scratch
    data_id = astra_data(VOL_GEOM_3D, 'volume', ndim=3)
    data_out = astra.data3d.get_shared(data_id)
    assert data_out.shape == (10, 20, 30)

    # From existing
    discr_dom = _discrete_domain(3, impl='numpy', device='cpu')
    data_in = discr_dom.element(np.ones((10, 20, 30), dtype='float32'))
    data_id = astra_data(VOL_GEOM_3D, 'volume', data=data_in)
    data_out = astra.data3d.get_shared(data_id)
    assert data_out.shape == (10, 20, 30)


PROJ_GEOM_2D = {
    'type': 'parallel',
    'DetectorCount': 15, 'DetectorWidth': 1.5,
    'ProjectionAngles': np.linspace(0, 2, 5)}

PROJ_GEOM_3D = {
    'type': 'parallel3d',
    'DetectorColCount': 15, 'DetectorRowCount': 25,
    'DetectorSpacingX': 1.5, 'DetectorSpacingY': 2.5,
    'ProjectionAngles': np.linspace(0, 2, 5)}


def test_parallel_2d_projector():
    """Create ASTRA 2D projectors."""
    # We can just test if it runs
    astra_projector('line', VOL_GEOM_2D, PROJ_GEOM_2D, ndim=2)
    astra_projector('linear', VOL_GEOM_2D, PROJ_GEOM_2D, ndim=2)

    with pytest.raises(ValueError):
        astra_projector('line_fanflat', VOL_GEOM_2D, PROJ_GEOM_2D, ndim=2)
    with pytest.raises(ValueError):
        astra_projector('linearcone', VOL_GEOM_2D, PROJ_GEOM_2D, ndim=2)


@pytest.mark.xfail(reason='3D CPU projectors not supported by ASTRA')
def test_parallel_3d_projector():
    """Create ASTRA 3D projectors."""
    # We can just test if it runs
    astra_projector('linear3d', VOL_GEOM_3D, PROJ_GEOM_3D, ndim=3)


@pytest.mark.skipif(not odl.applications.tomo.ASTRA_CUDA_AVAILABLE,
                    reason="ASTRA CUDA not available")
def test_parallel_3d_projector_gpu():
    """Create ASTRA 3D projectors on GPU."""
    astra_projector('cuda3d', VOL_GEOM_3D, PROJ_GEOM_3D, ndim=3)


def test_astra_algorithm():
    """Create ASTRA algorithm object."""
    direction = 'forward'
    ndim = 2
    impl = 'cpu'
    vol_id = astra_data(VOL_GEOM_2D, 'volume', ndim=ndim)
    sino_id = astra_data(PROJ_GEOM_2D, 'projection', ndim=ndim)
    proj_id = astra_projector('linear', VOL_GEOM_2D, PROJ_GEOM_2D, ndim=ndim)

    # Checks
    with pytest.raises(ValueError):
        astra_algorithm('none', ndim, vol_id, sino_id, proj_id, impl)
    with pytest.raises(ValueError):
        astra_algorithm(direction, 0, vol_id, sino_id, proj_id, impl)
    with pytest.raises(ValueError):
        astra_algorithm('none', ndim, vol_id, sino_id, proj_id, 'none')
    with pytest.raises(ValueError):
        astra_algorithm(
            'backward', ndim, vol_id, sino_id, proj_id=None, astra_impl='cpu'
        )
    alg_id = astra_algorithm(direction, ndim, vol_id, sino_id, proj_id, impl)
    astra.algorithm.delete(alg_id)

    # 2D CPU
    ndim = 2
    impl = 'cpu'
    for direction in {'forward', 'backward'}:
        alg_id = astra_algorithm(
            direction, ndim, vol_id, sino_id, proj_id, impl
        )
        astra.algorithm.delete(alg_id)


@pytest.mark.skipif(not odl.applications.tomo.ASTRA_CUDA_AVAILABLE,
                    reason="ASTRA cuda not available")
def test_astra_algorithm_gpu():
    """Create ASTRA algorithm object on GPU."""

    direction = 'forward'
    ndim = 2
    vol_id = astra_data(VOL_GEOM_2D, 'volume', ndim=ndim)
    rec_id = astra_data(VOL_GEOM_2D, 'volume', ndim=ndim)
    sino_id = astra_data(PROJ_GEOM_2D, 'projection', ndim=ndim)

    # 2D CUDA
    proj_id = astra_projector('cuda', VOL_GEOM_2D, PROJ_GEOM_2D, ndim=ndim)

    # 2D CUDA FP
    alg_id = astra_algorithm(
        'forward', ndim, vol_id, sino_id, proj_id=proj_id, astra_impl='cuda'
    )
    astra.algorithm.delete(alg_id)

    # 2D CUDA BP
    alg_id = astra_algorithm(
        'backward', ndim, rec_id, sino_id, proj_id=proj_id, astra_impl='cuda'
    )
    astra.algorithm.delete(alg_id)

    ndim = 3
    vol_id = astra_data(VOL_GEOM_3D, 'volume', ndim=ndim)
    sino_id = astra_data(PROJ_GEOM_3D, 'projection', ndim=ndim)
    proj_id = astra_projector('cuda3d', VOL_GEOM_3D, PROJ_GEOM_3D, ndim=ndim)

    with pytest.raises(NotImplementedError):
        astra_algorithm(
            direction, ndim, vol_id, sino_id, proj_id=proj_id, astra_impl='cpu'
        )

    for direction in {'forward', 'backward'}:
        astra_algorithm(
            direction, ndim, vol_id, sino_id, proj_id=proj_id, astra_impl='cuda'
        )


def test_geom_to_vec():
    """Create ASTRA projection geometry vectors using ODL geometries."""

    apart = odl.uniform_partition(0, 2 * np.pi, 5)
    dpart = odl.uniform_partition(-40, 40, 10)

    # Fanbeam flat
    src_rad = 10
    det_rad = 5
    geom_ff = odl.applications.tomo.FanBeamGeometry(apart, dpart, src_rad, det_rad)
    vec = odl.applications.tomo.astra_conebeam_2d_geom_to_vec(geom_ff)

    assert vec.shape == (apart.size, 6)

    # Circular cone flat
    dpart = odl.uniform_partition([-40, -3], [40, 3], (10, 5))
    geom_ccf = odl.applications.tomo.ConeBeamGeometry(apart, dpart, src_rad, det_rad)
    vec = odl.applications.tomo.astra_conebeam_3d_geom_to_vec(geom_ccf)
    assert vec.shape == (apart.size, 12)

    # Helical cone flat
    pitch = 1
    geom_hcf = odl.applications.tomo.ConeBeamGeometry(apart, dpart, src_rad, det_rad,
                                         pitch=pitch)
    vec = odl.applications.tomo.astra_conebeam_3d_geom_to_vec(geom_hcf)
    assert vec.shape == (apart.size, 12)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Phantoms typically used in transmission tomography."""

from __future__ import absolute_import, division, print_function

import numpy as np

from odl.core.discr import DiscretizedSpace
from odl.core.phantom.geometric import ellipsoid_phantom

__all__ = ('shepp_logan_ellipsoids', 'shepp_logan', 'forbild')


def _shepp_logan_ellipse_2d():
    """Return ellipse parameters for a 2d Shepp-Logan phantom.

    This assumes that the ellipses are contained in the square
    [-1, -1]x[-1, -1].
    """
    rad18 = np.deg2rad(18.0)
    #       value  axisx  axisy     x       y  rotation
    return [[2.00, .6900, .9200, 0.0000, 0.0000, 0],
            [-.98, .6624, .8740, 0.0000, -.0184, 0],
            [-.02, .1100, .3100, 0.2200, 0.0000, -rad18],
            [-.02, .1600, .4100, -.2200, 0.0000, rad18],
            [0.01, .2100, .2500, 0.0000, 0.3500, 0],
            [0.01, .0460, .0460, 0.0000, 0.1000, 0],
            [0.01, .0460, .0460, 0.0000, -.1000, 0],
            [0.01, .0460, .0230, -.0800, -.6050, 0],
            [0.01, .0230, .0230, 0.0000, -.6060, 0],
            [0.01, .0230, .0460, 0.0600, -.6050, 0]]


def _shepp_logan_ellipsoids_3d():
    """Return ellipsoid parameters for a 3d Shepp-Logan phantom.

    This assumes that the ellipses are contained in the cube
    [-1, -1, -1]x[1, 1, 1].
    """
    rad18 = np.deg2rad(18.0)
    #       value  axisx  axisy  axisz,  x        y      z    rotation
    return [[2.00, .6900, .9200, .810, 0.0000, 0.0000, 0.00, 0.0, 0, 0],
            [-.98, .6624, .8740, .780, 0.0000, -.0184, 0.00, 0.0, 0, 0],
            [-.02, .1100, .3100, .220, 0.2200, 0.0000, 0.00, -rad18, 0, 0],
            [-.02, .1600, .4100, .280, -.2200, 0.0000, 0.00, rad18, 0, 0],
            [0.01, .2100, .2500, .410, 0.0000, 0.3500, 0.00, 0.0, 0, 0],
            [0.01, .0460, .0460, .050, 0.0000, 0.1000, 0.00, 0.0, 0, 0],
            [0.01, .0460, .0460, .050, 0.0000, -.1000, 0.00, 0.0, 0, 0],
            [0.01, .0460, .0230, .050, -.0800, -.6050, 0.00, 0.0, 0, 0],
            [0.01, .0230, .0230, .020, 0.0000, -.6060, 0.00, 0.0, 0, 0],
            [0.01, .0230, .0460, .020, 0.0600, -.6050, 0.00, 0.0, 0, 0]]


def _modified_shepp_logan_ellipsoids(ellipsoids):
    """Modify ellipsoids to give the modified Shepp-Logan phantom.

    Works for both 2d and 3d.
    """
    intensities = [1.0, -0.8, -0.2, -0.2, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    # Add minimal numbers to ensure that the result is nowhere negative.
    # This is needed due to numerical issues.
    intensities[2] += 5e-17
    intensities[3] += 5e-17

    assert len(ellipsoids) == len(intensities)

    for ellipsoid, intensity in zip(ellipsoids, intensities):
        ellipsoid[0] = intensity


def shepp_logan_ellipsoids(ndim, modified=False):
    """Ellipsoids for the standard Shepp-Logan phantom in 2 or 3 dimensions.

    Parameters
    ----------
    ndim : {2, 3}
        Dimension of the space the ellipsoids should be in.
    modified : bool, optional
        True if the modified Shepp-Logan phantom should be given.
        The modified phantom has greatly amplified contrast to aid
        visualization.

    See Also
    --------
    odl.core.phantom.geometric.ellipsoid_phantom :
        Function for creating arbitrary ellipsoids phantoms
    shepp_logan : Create a phantom with these ellipsoids

    References
    ----------
    .. _Shepp-Logan phantom: https://en.wikipedia.org/wiki/Shepp-Logan_phantom
    """
    if ndim == 2:
        ellipsoids = _shepp_logan_ellipse_2d()
    elif ndim == 3:
        ellipsoids = _shepp_logan_ellipsoids_3d()
    else:
        raise ValueError('dimension not 2 or 3, no phantom available')

    if modified:
        _modified_shepp_logan_ellipsoids(ellipsoids)

    return ellipsoids


def shepp_logan(space, modified=False, min_pt=None, max_pt=None):
    """Standard Shepp-Logan phantom in 2 or 3 dimensions.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Space in which the phantom is created, must be 2- or 3-dimensional.
        If ``space.shape`` is 1 in an axis, a corresponding slice of the
        phantom is created.
    modified : `bool`, optional
        True if the modified Shepp-Logan phantom should be given.
        The modified phantom has greatly amplified contrast to aid
        visualization.
    min_pt, max_pt : array-like, optional
        If provided, use these vectors to determine the bounding box of the
        phantom instead of ``space.min_pt`` and ``space.max_pt``.
        It is currently required that ``min_pt >= space.min_pt`` and
        ``max_pt <= space.max_pt``, i.e., shifting or scaling outside the
        original space is not allowed.

        Providing one of them results in a shift, e.g., for ``min_pt``::

            new_min_pt = min_pt
            new_max_pt = space.max_pt + (min_pt - space.min_pt)

        Providing both results in a scaled version of the phantom.

    See Also
    --------
    forbild : Similar phantom but with more complexity. Only supports 2d.
    odl.core.phantom.geometric.defrise : Geometry test phantom
    shepp_logan_ellipsoids : Get the parameters that define this phantom
    odl.core.phantom.geometric.ellipsoid_phantom :
        Function for creating arbitrary ellipsoid phantoms

    References
    ----------
    .. _Shepp-Logan phantom: https://en.wikipedia.org/wiki/Shepp-Logan_phantom
    """
    ellipsoids = shepp_logan_ellipsoids(space.ndim, modified)
    return ellipsoid_phantom(space, ellipsoids, min_pt, max_pt)


def _analytical_forbild_phantom(resolution, ear):
    """Analytical description of FORBILD phantom.

    Parameters
    ----------
    resolution : bool
        If ``True``, insert a small resolution test pattern to the left.
    ear : bool
        If ``True``, insert an ear-like structure to the right.
    """
    sha = 0.2 * np.sqrt(3)
    y016b = -14.294530834372887
    a16b = 0.443194085308632
    b16b = 3.892760834372886

    E = [[-4.7, 4.3, 1.79989, 1.79989, 0, 0.010, 0],  # 1
         [4.7, 4.3, 1.79989, 1.79989, 0, 0.010, 0],  # 2
         [-1.08, -9, 0.4, 0.4, 0, 0.0025, 0],  # 3
         [1.08, -9, 0.4, 0.4, 0, -0.0025, 0],  # 4
         [0, 0, 9.6, 12, 0, 1.800, 0],  # 5
         [0, 8.4, 1.8, 3.0, 0, -1.050, 0],  # 7
         [1.9, 5.4, 0.41633, 1.17425, -31.07698, 0.750, 0],  # 8
         [-1.9, 5.4, 0.41633, 1.17425, 31.07698, 0.750, 0],  # 9
         [-4.3, 6.8, 1.8, 0.24, -30, 0.750, 0],  # 10
         [4.3, 6.8, 1.8, 0.24, 30, 0.750, 0],  # 11
         [0, -3.6, 1.8, 3.6, 0, -0.005, 0],  # 12
         [6.39395, -6.39395, 1.2, 0.42, 58.1, 0.005, 0],  # 13
         [0, 3.6, 2, 2, 0, 0.750, 4],  # 14
         [0, 9.6, 1.8, 3.0, 0, 1.800, 4],  # 15
         [0, 0, 9.0, 11.4, 0, 0.750, 3],  # 16a
         [0, y016b, a16b, b16b, 0, 0.750, 1],  # 16b
         [0, 0, 9.0, 11.4, 0, -0.750, ear],  # 6
         [9.1, 0, 4.2, 1.8, 0, 0.750, 1]]  # R_ear
    E = np.array(E)

    # generate the air cavities in the right ear
    cavity1 = np.arange(8.8, 5.6, -0.4)[:, None]
    cavity2 = np.zeros([9, 1])
    cavity3_7 = np.ones([53, 1]) * [0.15, 0.15, 0, -1.800, 0]

    for j in range(1, 4):
        kj = 8 - 2 * int(np.floor(j / 3))
        dj = 0.2 * int(np.mod(j, 2))

        cavity1 = np.vstack((cavity1,
                             cavity1[0:kj] - dj,
                             cavity1[0:kj] - dj))
        cavity2 = np.vstack((cavity2,
                             j * sha * np.ones([kj, 1]),
                             -j * sha * np.ones([kj, 1])))

    E_cavity = np.hstack((cavity1, cavity2, cavity3_7))

    # generate the left ear (resolution pattern)
    x0 = -7.0
    y0 = -1.0
    d0_xy = 0.04

    d_xy = [0.0357, 0.0312, 0.0278, 0.0250]
    ab = 0.5 * np.ones([5, 1]) * d_xy
    ab = ab.T.ravel()[:, None] * np.ones([1, 4])
    abr = ab.T.ravel()[:, None]

    leftear4_7 = np.hstack([abr, abr, np.ones([80, 1]) * [0, 0.75, 0]])

    x00 = np.zeros([0, 1])
    y00 = np.zeros([0, 1])
    for i in range(1, 5):
        y00 = np.vstack((y00,
                         (y0 + np.arange(0, 5) * 2 * d_xy[i - 1])[:, None]))
        x00 = np.vstack((x00,
                         (x0 + 2 * (i - 1) * d0_xy) * np.ones([5, 1])))

    x00 = x00 * np.ones([1, 4])
    x00 = x00.T.ravel()[:, None]
    y00 = np.vstack([y00, y00 + 12 * d0_xy,
                     y00 + 24 * d0_xy, y00 + 36 * d0_xy])

    leftear = np.hstack([x00, y00, leftear4_7])
    C = [[1.2, 1.2, 0.27884, 0.27884, 0.60687, 0.60687, 0.2,
          0.2, -2.605, -2.605, -10.71177, y016b + 10.71177, 8.88740, -0.21260],
         [0, 180, 90, 270, 90, 270, 0,
          180, 15, 165, 90, 270, 0, 0]]
    C = np.array(C)

    if not resolution and not ear:
        phantomE = E[:17, :]
        phantomC = C[:, :12]
    elif not resolution and ear:
        phantomE = np.vstack([E, E_cavity])
        phantomC = C
    elif resolution and not ear:
        phantomE = np.vstack([leftear, E[:17, :]])
        phantomC = C[:, :12]
    else:
        phantomE = np.vstack([leftear, E, E_cavity])
        phantomC = C

    return phantomE, phantomC


def forbild(space, resolution=False, ear=True, value_type='density',
            scale='auto'):
    """Standard FORBILD phantom in 2 dimensions.

    The FORBILD phantom is intended for testing CT algorithms and is intended
    to be similar to a human head.

    The phantom is defined using the following materials:

    =========================  =====  ================
    Material                   Index  Density (g/cm^3)
    =========================  =====  ================
    Air                        0      0.0000
    Cerebrospinal fluid (CSF)  1      1.0450
    Small less dense sphere    2      1.0475
    Brain                      3      1.0500
    Small more dense sphere    4      1.0525
    Blood                      5      1.0550
    Eyes                       6      1.0600
    Bone                       7      1.8000
    =========================  =====  ================

    Parameters
    ----------
    space : `DiscretizedSpace`
        The space in which the phantom should be corrected. Needs to be two-
        dimensional.
    resolution : bool, optional
        If ``True``, insert a small resolution test pattern to the left.
    ear : bool, optional
        If ``True``, insert an ear-like structure to the right.
    value_type : {'density', 'materials'}, optional
        The format the phantom should be given in.
        'density' returns floats in the range [0, 1.8] (g/cm^3)
        'materials' returns indices in the range [0, 7].
    scale : {'auto', 'cm', 'meters', 'mm'}, optional
        Controls how ``space`` should be rescaled to fit the definition of
        the forbild phantom, which is defined on the square
        [-12.8, 12.8] x [-12.8, 12.8] cm.

        * ``'auto'`` means that space is rescaled to fit exactly. The space is
          also centered at [0, 0].
        * ``'cm'`` means the dimensions of the space should be used as-is.
        * ``'m'`` means all dimensions of the space are multiplied by 100.
        * ``'mm'`` means all dimensions of the space are divided by 10.

    Returns
    -------
    forbild : ``space``-element
        FORBILD phantom discretized by ``space``.

    See Also
    --------
    shepp_logan : A simpler phantom for similar purposes, also working in 3d.

    References
    ----------
    .. _FORBILD phantom: www.imp.uni-erlangen.de/phantoms/head/head.html
    .. _algorithm: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3426508/
    """
    def transposeravel(arr):
        """Implement MATLAB's ``transpose(arr(:))``."""
        return arr.T.ravel()

    if not isinstance(space, DiscretizedSpace):
        raise TypeError('`space` must be a `DiscretizedSpace`')
    if space.ndim != 2:
        raise TypeError('`space` must be two-dimensional')

    scale, scale_in = str(scale).lower(), scale
    value_type, value_type_in = str(value_type).lower(), value_type

    # Create analytic description of phantom
    phantomE, phantomC = _analytical_forbild_phantom(resolution, ear)

    # Rescale points to the default grid.
    # The forbild phantom is defined on [-12.8, 12.8] x [-12.8, 12.8]
    xcoord, ycoord = space.points().T
    if scale == 'auto':
        xcoord = ((xcoord - space.min_pt[0]) /
                  (space.max_pt[0] - space.min_pt[0]))
        xcoord = 25.8 * xcoord - 12.8
        ycoord = ((ycoord - space.min_pt[1]) /
                  (space.max_pt[1] - space.min_pt[1]))
        ycoord = 25.8 * ycoord - 12.8
    elif scale == 'cm':
        pass  # dimensions already correct.
    elif scale == 'm':
        xcoord *= 100.0
        ycoord *= 100.0
    elif scale == 'mm':
        xcoord /= 10.0
        ycoord /= 10.0
    else:
        raise ValueError('unknown `scale` {}'.format(scale_in))

    # Compute the phantom values in each voxel
    image = np.zeros(space.size)
    nclipinfo = 0
    for k in range(phantomE.shape[0]):
        # Handle elliptic bounds
        Vx0 = np.array([transposeravel(xcoord) - phantomE[k, 0],
                        transposeravel(ycoord) - phantomE[k, 1]])
        D = np.array([[1 / phantomE[k, 2], 0],
                      [0, 1 / phantomE[k, 3]]])
        phi = np.deg2rad(phantomE[k, 4])
        Q = np.array([[np.cos(phi), np.sin(phi)],
                      [-np.sin(phi), np.cos(phi)]])
        f = phantomE[k, 5]
        nclip = int(phantomE[k, 6])
        equation1 = np.sum(D.dot(Q).dot(Vx0) ** 2, axis=0)
        i = (equation1 <= 1.0)

        # Handle clipping surfaces
        for _ in range(nclip):  # note: nclib can be 0
            d = phantomC[0, nclipinfo]
            psi = np.deg2rad(phantomC[1, nclipinfo])
            equation2 = np.array([np.cos(psi), np.sin(psi)]).dot(Vx0)
            i &= (equation2 < d)
            nclipinfo += 1

        image[i] += f

    if value_type == 'materials':
        materials = np.zeros(space.size, dtype=space.dtype)
        # csf
        materials[(image > 1.043) & (image <= 1.047)] = 1
        # less_dense_sphere
        materials[(image > 1.047) & (image <= 1.048)] = 2
        # brain
        materials[(image > 1.048) & (image <= 1.052)] = 3
        # denser_sphere
        materials[(image > 1.052) & (image <= 1.053)] = 4
        # blood
        materials[(image > 1.053) & (image <= 1.058)] = 5
        # eye
        materials[(image > 1.058) & (image <= 1.062)] = 6
        # Bone
        materials[image > 1.75] = 7

        return space.element(materials.reshape(space.shape))
    elif value_type == 'density':
        return space.element(image.reshape(space.shape))
    else:
        raise ValueError('unknown `value_type` {}'.format(value_type_in))


if __name__ == '__main__':
    # Show the phantoms
    import odl
    from odl.core.util.testutils import run_doctests

    # 2D
    discr = odl.uniform_discr([-1, -1], [1, 1], [1000, 1000])
    shepp_logan(discr, modified=True).show('shepp_logan 2d modified=True')
    shepp_logan(discr, modified=False).show('shepp_logan 2d modified=False')
    forbild(discr).show('FORBILD 2d', clim=[1.035, 1.065])
    forbild(discr, value_type='materials').show('FORBILD 2d materials')

    # 3D
    discr = odl.uniform_discr([-1, -1, -1], [1, 1, 1], [300, 300, 300])
    shepp_logan(discr, modified=True).show('shepp_logan 3d modified=True')
    shepp_logan(discr, modified=False).show('shepp_logan 3d modified=False')

    # Run also the doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the forward-backward solver."""

from __future__ import division
import pytest

import odl
from odl.solvers import forward_backward_pd
from odl.core.util.testutils import all_almost_equal, noise_element

# Places for the accepted error when comparing results
HIGH_ACCURACY = 8
LOW_ACCURACY = 4


def test_forward_backward_input_handling():
    """Test to see that input is handled correctly."""

    space1 = odl.uniform_discr(0, 1, 10)

    lin_ops = [odl.ZeroOperator(space1), odl.ZeroOperator(space1)]
    g = [odl.functionals.ZeroFunctional(space1),
         odl.functionals.ZeroFunctional(space1)]
    f = odl.functionals.ZeroFunctional(space1)
    h = odl.functionals.ZeroFunctional(space1)

    # Check that the algorithm runs. With the above operators, the algorithm
    # returns the input.
    x0 = noise_element(space1)
    x = x0.copy()
    niter = 3

    forward_backward_pd(x, f, g, lin_ops, h, tau=1.0,
                        sigma=[1.0, 1.0], niter=niter)

    assert x == x0

    # Testing that sizes needs to agree:
    # Too few sigma_i:s
    with pytest.raises(ValueError):
        forward_backward_pd(x, f, g, lin_ops, h, tau=1.0,
                            sigma=[1.0], niter=niter)

    # Too many operators
    g_too_many = [odl.functionals.ZeroFunctional(space1),
                  odl.functionals.ZeroFunctional(space1),
                  odl.functionals.ZeroFunctional(space1)]
    with pytest.raises(ValueError):
        forward_backward_pd(x, f, g_too_many, lin_ops, h,
                            tau=1.0, sigma=[1.0, 1.0], niter=niter)

    # Test for correct space
    space2 = odl.uniform_discr(1, 2, 10)
    x = noise_element(space2)
    with pytest.raises(ValueError):
        forward_backward_pd(x, f, g, lin_ops, h, tau=1.0,
                            sigma=[1.0, 1.0], niter=niter)


def test_forward_backward_basic():
    """Test for the forward-backward solver by minimizing ||x||_2^2.

    The general problem is of the form

        ``min_x f(x) + sum_i g_i(L_i x) + h(x)``

    and here we take f(x) = g(x) = 0, h(x) = ||x||_2^2 and L is the
    zero-operator.
    """

    space = odl.rn(10)

    lin_ops = [odl.ZeroOperator(space)]
    g = [odl.functionals.ZeroFunctional(space)]
    f = odl.functionals.ZeroFunctional(space)
    h = odl.functionals.L2NormSquared(space)

    x = noise_element(space)
    x_global_min = space.zero()

    forward_backward_pd(x, f, g, lin_ops, h, tau=0.5,
                        sigma=[1.0], niter=10)

    assert all_almost_equal(x, x_global_min, ndigits=HIGH_ACCURACY)


def test_forward_backward_with_lin_ops():
    """Test for the forward-backward solver with linear operatros.

    The test is done by minimizing ||x - b||_2^2 + ||alpha * x||_2^2. The
    general problem is of the form

        ``min_x f(x) + sum_i g_i(L_i x) + h(x)``

    and here we take f = 0, g = ||.||_2^2, L = alpha * IndentityOperator,
    and h = ||. - b||_2^2.
    """

    space = odl.rn(10)
    alpha = 0.1
    b = noise_element(space)

    lin_ops = [alpha * odl.IdentityOperator(space)]
    g = [odl.functionals.L2NormSquared(space)]
    f = odl.functionals.ZeroFunctional(space)

    # Gradient of two-norm square
    h = odl.functionals.L2NormSquared(space).translated(b)

    x = noise_element(space)

    # Explicit solution: x_hat = (I^T * I + (alpha*I)^T * (alpha*I))^-1 * (I*b)
    x_global_min = b / (1 + alpha ** 2)

    forward_backward_pd(x, f, g, lin_ops, h, tau=0.5,
                        sigma=[1.0], niter=20)

    assert all_almost_equal(x, x_global_min, ndigits=LOW_ACCURACY)


def test_forward_backward_with_li():
    """Test for the forward-backward solver with infimal convolution.

    The test is done by minimizing the functional ``(g @ l)(x)``, where

        ``(g @ l)(x) = inf_y { g(y) + l(x-y) }``,

    g is the indicator function on [-3, -1], and l(x) = 1/2||x||_2^2.
    The optimal solution to this problem is given by x in [-3, -1].
    """
    # Parameter values for the box constraint
    upper_lim = -1
    lower_lim = -3

    space = odl.rn(1)

    lin_op = odl.IdentityOperator(space)
    lin_ops = [lin_op]
    g = [odl.functionals.IndicatorBox(space, lower=lower_lim, upper=upper_lim)]
    f = odl.functionals.ZeroFunctional(space)
    h = odl.functionals.ZeroFunctional(space)
    l = [0.5 * odl.functionals.L2NormSquared(space)]

    # Creating an element not to far away from [-3,-1], in order to converge in
    # a few number of iterations.
    x = space.element(10)

    forward_backward_pd(x, f, g, lin_ops, h, tau=0.5,
                        sigma=[1.0], niter=20, l=l)

    assert lower_lim - 10 ** (-LOW_ACCURACY) <= x[0]
    assert x[0] <= upper_lim + 10 ** (-LOW_ACCURACY)


def test_forward_backward_with_li_and_h():
    """Test for the forward-backward solver with infimal convolution.

    The test is done by minimizing the functional ``(g @ l)(x) + h(x)``, where

        ``(g @ l)(x) = inf_y { g(y) + l(x-y) }``,

    g is the indicator function on [-3, -1], and l(x) = h(x) = 1/2||x||_2^2.
    The optimal solution to this problem is given by x = -0.5.
    """

    # Parameter values for the box constraint
    upper_lim = -1
    lower_lim = -3

    space = odl.rn(1)

    lin_ops = [odl.IdentityOperator(space)]
    g = [odl.functionals.IndicatorBox(space, lower=lower_lim, upper=upper_lim)]
    f = odl.functionals.ZeroFunctional(space)
    h = 0.5 * odl.functionals.L2NormSquared(space)
    l = [0.5 * odl.functionals.L2NormSquared(space)]

    # Creating an element not to far away from -0.5, in order to converge in
    # a few number of iterations.
    x = space.element(10)

    forward_backward_pd(x, f, g, lin_ops, h, tau=0.5,
                        sigma=[1.0], niter=20, l=l)

    expected_result = -0.5
    assert x[0] == pytest.approx(expected_result, rel=10 ** -LOW_ACCURACY)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)
